﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace NavigationController {
	public class NavigationData {

		public const double aDMultiplier = 3.3 / 65472; //see coridium AD function documentation
		public const double resistorDivider = 1.0 / 6.0; //resistor dividers on system 
		//battery and navigation battery (R1 = 10K, R2 = 50K)
		public const double accelerometerNull = 1.66;
		public readonly double magneticDeclination;
		public const double vPerG = 0.333;
		public const double xOffset = 0.13;
		public const double yOffset = 0.07;
		public const double zOffset = 0.14;

		string _error = "";
		public string Error {
			get { return _error; }
		}

		string _navigationBuffer = "";
		public string NavigationBuffer {
			get { return _navigationBuffer; }
			set { _navigationBuffer = value; }
		}

		//gets the average x-axis acceleration
		public double XAccAverage {
			get { return getRealAcc(_xAccList.Average(), xOffset); }
		}

		//general purpose accelerometer raw value to real value conversion
		public double getRealAcc(double acc, double offset) {
			return ((aDMultiplier * acc - accelerometerNull) / vPerG) + offset;
		}

		double _xAcc = 0;

		public double XAcc {
			get { return _xAcc; }
			set { _xAcc = value; }
		}
		List<double> _xAccList;
		public List<double> XAccList {
			get { return _xAccList; }
		}

		public double YAccAverage {
			get { return getRealAcc(_yAccList.Average(), yOffset); }
		}

		double _yAcc = 0;

		public double YAcc {
			get { return _yAcc; }
			set { _yAcc = value; }
		}
		List<double> _yAccList;
		public List<double> YAccList {
			get { return _yAccList; }
		}

		public double ZAccAverage {
			get { return getRealAcc(_zAccList.Average(), zOffset); }
		}

		double _zAcc = 0;

		public double ZAcc {
			get { return _zAcc; }
			set { _zAcc = value; }
		}
		List<double> _zAccList;
		public List<double> ZAccList {
			get { return _zAccList; }
		}

		public double LeftIrAverage {
			get { return getIrCm(_leftIrList.Average()); }
		}

		public double getIrCm(double voltage) {
			/*
			cubic regression conversion formula
			for turning voltage into centimeters
			for the sharp 2Y0A02 IR rangefinder
			y = ax^3 + bx^2 + cx + d
			a = -37.22043895
			b = 211.4582599
			c = -399.7906762
			d = 285.2892273
			 */
			double x = aDMultiplier * voltage;
			double a = -37.22043895, b = 211.4582599, c = -399.7906762, d = 285.2892273;
			return a * Math.Pow(x, 3) + b * Math.Pow(x, 2) + c * x + d;
			//return voltage;
		}

		double _leftIr = 0;

		public double LeftIr {
			get { return _leftIr; }
			set { _leftIr = value; }
		}
		List<double> _leftIrList;
		public List<double> LeftIrList {
			get { return _leftIrList; }
		}

		public double RightIrAverage {
			get { return getIrCm(_rightIrList.Average()); }
		}

		double _rightIr = 0;

		public double RightIr {
			get { return _rightIr; }
			set { _rightIr = value; }
		}
		List<double> _rightIrList;
		public List<double> RightIrList {
			get { return _rightIrList; }
		}

		public double CompassHeadingAverage {
			get {
				if (_compassHeadList.Count > 0)
					return getRealCompassHeading(_compassHeadList.Average());
				else
					return 0;
			}
		}

		public double getRealCompassHeading(double heading) {
			heading /= 10.0;
			if (heading >= 360.0)
				heading = heading - 360.0;
			return heading;
		}


		List<double> _compassFilterList;

		double _compassHeading = 0;

		public double CompassHeading {
			get { return _compassHeading; }
			set { _compassHeading = value; }
		}
		List<double> _compassHeadList;
		public List<double> CompassHeadList {
			get { return _compassHeadList; }
		}

		public double SystemBatteryAverage {
			get { return getRealVoltage(_systemBattList.Average()); }
		}

		public double getRealVoltage(double voltage) {
			return aDMultiplier * voltage / resistorDivider;
		}

		double _systemBattery = 0;

		public double SystemBattery {
			get { return _systemBattery; }
			set { _systemBattery = value; }
		}
		List<double> _systemBattList;
		public List<double> SystemBattList {
			get { return _systemBattList; }
		}

		public double PlatformBatteryAverage {
			get { return getRealVoltage(_platformBattList.Average()); }
		}

		double _platformBattery = 0;

		public double PlatformBattery {
			get { return _platformBattery; }
			set { _platformBattery = value; }
		}
		List<double> _platformBattList;
		public List<double> PlatformBattList {
			get { return _platformBattList; }
		}

		string _status = "not initialized";

		public string Status {
			get { return _status; }
			set { _status = value; }
		}

		public NavigationData(double magDec) {
			magneticDeclination = magDec;
			_xAccList = new List<double>();
			_yAccList = new List<double>();
			_zAccList = new List<double>();
			_leftIrList = new List<double>();
			_rightIrList = new List<double>();
			//use to keep a history of the compass 
			_compassFilterList = new List<double>();
			_compassHeadList = new List<double>();
			_systemBattList = new List<double>();
			_platformBattList = new List<double>();
		}
	}
}
