﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using UtmConvert;

namespace system_controller {
    class RaceStrategy {

		public delegate void updateUiEventHandler();
		public event updateUiEventHandler updateUiEvent;

		SystemData _data;
		int _pointCount = 0;
		public int PointCount {
			get { return _pointCount; }
		}

		UtmConvert.UtmPoint _currentLocationVector = new UtmPoint();
		public UtmConvert.UtmPoint CurrentVectorLocation {
			get { return _currentLocationVector; }
		}

		UtmConvert.UtmPoint _currentLocationGps = new UtmPoint();
		public UtmConvert.UtmPoint CurrentGpsLocation {
			get { return _currentLocationGps; }
		}

		public readonly double steeringKp;
		//the distance the robot must be from the cone for it to start using
		//the vision system to look for it
		public readonly double coneDistanceThreshold; //meters
		//the distance the robot must be from a dummy waypoint 
		//in order to increment the waypoint counter
		public readonly double dummyDistanceThreshold; //meters
		double _waypointDistanceVector = 0;
		public double WaypointDistanceVector {
			get { return _waypointDistanceVector; }
		}
		//distance in meters between the robot and the waypoint 
		//according to gps measurments
		double _waypointDistanceGps = 0;
		public double WaypointDistanceGps {
			get { return _waypointDistanceGps; }
		}
		//the angular difference between the robots
		//vector heading and position and theta
		double _crossTrackErrorVector = 0;
		public double CrossTrackErrorVector {
			get { return _crossTrackErrorVector; }
		}
		//the angular difference between the robots
		//gps heading and position and theta
		double _crossTrackErrorGps = 0;
		public double CrossTrackErrorGps {
			get { return _crossTrackErrorGps; }
		}
		//
		string _turnDirection = "null"; //left, right
		public string TurnDirection {
			get { return _turnDirection; }
		}
		//theta according to vector navigation
		double _thetaVector = 0;
		public double ThetaVector {
			get { return _thetaVector; }
		}

		//modified compass heading 
		double _offsetCompassHeading;
		public double OffsetCompassHeading {
			get { return _offsetCompassHeading; }
		}

		bool _arrivedAtWaypoint = true;
		public bool ArrivedAtWaypoint {
			get { return _arrivedAtWaypoint; }
		}

		//the direction of the waypoint according to the gps location
		double _thetaGps = 0;
		//the x axis delta between the waypoint and the robot
		double _xDeltaVector = 0;
		//the y axis delta between the waypoint and the robot
		double _yDeltaVector = 0;
		double _xDeltaGps = 0;
		//the y axis delta between the waypoint and the robot
		double _yDeltaGps = 0;
		//tells if the robot has gotten lost 
		//while using vector navigation
		//if below 95 then start relying on GPS
		int _vectorPositionConfidence = 100;
		int _gpsPositionConfidence = 0;

		int _viewAttempts = 0;
		int _coneViewCount = 0; //confidence level that it is a cone
		bool _touchedCone = false;
		//tells whether the robot is looking for the cone
		bool _lookingForCone = false;
		public bool LookingForCone {
			get { return _lookingForCone; }
		}

		VectorCalculator _vCalc;
		double _previousHeading = 0;
		double _previousDistance = 0;
		const int _vectorGenerationThreshold = 2;

		double navigationSpeed = 0;
		double platformSpeed = 0;
		double speed = 0;
		readonly int minimumSpeed;
		readonly int maximumSpeed;

		double steering = 0;
		DataLogger _dataLog;
		EventLogger _eventLog;

		public RaceStrategy(ref SystemData data, double steerKp
			, double coneDistance, double dummyDistance, int minSpeed, int maxSpeed) {
			steeringKp = steerKp;
			coneDistanceThreshold = coneDistance;
			dummyDistanceThreshold = dummyDistance;
			minimumSpeed = minSpeed;
			maximumSpeed = maxSpeed;
			_currentLocationVector = new UtmConvert.UtmPoint();
			_dataLog = new DataLogger();
			_eventLog = new EventLogger();
			_dataLog.openFile();
			_eventLog.openFile();
			_eventLog.write(DateTime.Now.ToString() + ": race started\r\n");
			//print column headings for the data logger
			_dataLog.write(
				//Race Strategy Data
				"Date/Time\tPointCount\tCurrent Location Vector X\tCurrent Location Vector Y"
				+ "\tCurrent Location Gps X\tCurrent Location Gps Y\tWaypoint Distance Vector (M)"
				+ "\tWaypoint Distance Gps (M)\tCross Track Error Vector (Deg)\tCross Track Error GPS (Deg)"
				+ "\tTurn Direction\tTheta Vector (Deg)\tTheta GPS (Deg)\tLooking For Cone"
				+ "\tConeSearch Pattern Number\tView Attempts\tCone View Count\tTouched Cone"
				+ "\tArrived At Waypoint\tNavigation Speed\tPlatform Speed\tSpeed\tStopping"
				//Platform Data
				+ "\tEncoder Counts\tMeters\tCounts Per Second\tSpeed"
				+ "\tThrottle\tSteering\tReversing\tLeft Sonar (cm)\tCenter Sonar (cm)"
				+ "\tRight Sonar (cm)\tLeft Bumper\tRight Bumper\tManual Override"
				+ "\tAuto Avoid\tIR Instruction\tPlatform Status"
				//Navigation Data
				+ "\tCompass Heading (Deg)\tLeft IR (cm)\tRight IR (cm)"
				+ "\tX Axis Acc (G)\tY Axis Acc (G)\tZ Axis Acc (G)"
				+ "\tPlatform Battery (V)\tSystem Battery (V)\tNavigation Status"
				//GPS Data
				+ "\tGPS Quality Indicator\tSatelites in View\tLatitude\tNorth or South"
				+ "\tLongitude\tEast or West\tAntenna Altitude\tUnits of Antenna Altitude"
				+ "\tGeoidal Seperation\tUnits of Geoidal Seperation\tHorizontal Dilution of Precision"
				+ "\tGPS Time\tSpeed over Ground (knots)\tTrack Made Good"
				+ "\tMagnetic Variation\tEast or West Meters\r\n");
			_data = data;
			_data.initPlatform();
			_data.initNavigation();
			_data.initGps();
			_data.initVision();
			calculateOffsetCompassHeading();
			_previousHeading = _offsetCompassHeading;
			_previousDistance = _data.Platform.Data.Meters;
			_data.Navigation.updateDataEvent += navigateToWaypoint;
			_data.Navigation.writeToNavigationController(
				NavigationController.NavigationCommands
				.cmdArray[(int)NavigationController
				.NavigationCommands.commands.systemControllerUp]);
		}

		//update the robots current location
		public void updateCurrentLocation() {
			if (_arrivedAtWaypoint) {
				_currentLocationVector = new UtmPoint(new Point(
					_data.Parser.UtmPointSetList[0].Points[_pointCount].Point.X
					, _data.Parser.UtmPointSetList[0].Points[_pointCount].Point.Y)
					, _data.Parser.UtmPointSetList[0].Points[_pointCount].Zone
					, _data.Parser.UtmPointSetList[0].Points[_pointCount].Note
					, _data.Parser.UtmPointSetList[0].Points[_pointCount].Handle);
				_vCalc = new VectorCalculator(_currentLocationVector);
				//only increment this in one place
				if (_pointCount + 1 < _data.Parser.UtmPointSetList[0].Points.Count) {
					_eventLog.write(DateTime.Now.ToString() + ": arrived at waypoint "
						+ _pointCount.ToString() + ", " + 
						_data.Parser.UtmPointSetList[0].Points[_pointCount].Handle + ", " 
						+ _data.Parser.UtmPointSetList[0].Points[_pointCount].Note + "\r\n");
					++_pointCount;
					_arrivedAtWaypoint = false;
				} else _stopping = true;
			}
			if (_data.Gps.Gga.Data.QualityIndicator > 0) {
				_gpsPositionConfidence = _data.Gps.Gga.Data.SatelitesInView;
				_data.ConvertUtm.convertLatLonToUtm(ConvertDegRad.getRadians(
					_data.Gps.Gga.Data.LatDeg, _data.Gps.Gga.Data.LatMin
					, _data.Gps.Gga.Data.NorthOrSouth)
					, ConvertDegRad.getRadians(_data.Gps.Gga.Data.LonDeg
					, _data.Gps.Gga.Data.LonMin, _data.Gps.Gga.Data.EastOrWest));
				_currentLocationGps = new UtmPoint(new Point(_data.ConvertUtm.Easting
					, _data.ConvertUtm.Northing), _data.ConvertUtm.Zone, "", "");
			}
			//use vector addition to update location
			calculateOffsetCompassHeading();
			if (_previousDistance != _data.Platform.Data.Meters) {
				_currentLocationVector = _vCalc.getPointFromVector(_currentLocationVector
					, _data.Platform.Data.Meters - _previousDistance
					, (_previousHeading + _offsetCompassHeading) / 2);
				_previousHeading = _offsetCompassHeading;
				_previousDistance = _data.Platform.Data.Meters;
			}
		}

		public void checkArrivedAtWaypointWaypoint() {
			//calculate x and y deltas
			_xDeltaVector = _data.Parser.UtmPointSetList[0]
				.Points[_pointCount].Point.X - _currentLocationVector.Point.X;
			_yDeltaVector = _data.Parser.UtmPointSetList[0]
				.Points[_pointCount].Point.Y - _currentLocationVector.Point.Y;
			//update vector position
			_waypointDistanceVector = Math.Sqrt(
				Math.Pow(_xDeltaVector, 2) + Math.Pow(_yDeltaVector, 2));
			if (_data.Gps.Gga.Data.QualityIndicator > 0) {
				//update gps position
				_xDeltaGps = _data.Parser.UtmPointSetList[0]
					.Points[_pointCount].Point.X - _currentLocationGps.Point.X;
				_yDeltaGps = _data.Parser.UtmPointSetList[0]
					.Points[_pointCount].Point.Y - _currentLocationGps.Point.Y;
				_waypointDistanceGps = Math.Sqrt(
					Math.Pow(_xDeltaGps, 2) + Math.Pow(_yDeltaGps, 2));
			}
			//if(cone waypoint then must be touching cone)
			//warning: must be able to navigate to cone 
			//even if it is immediately close by
			if ((_vectorPositionConfidence > 95 && _data.Parser.UtmPointSetList[0]
				.Points[_pointCount].Note == "cone"
				&& _waypointDistanceVector <= coneDistanceThreshold)
				|| (_gpsPositionConfidence > 16 && _data.Parser.UtmPointSetList[0]
					.Points[_pointCount].Note == "cone"
				&& _waypointDistanceGps <= coneDistanceThreshold)) {
				//this sets arrivedAtWaypoint to true if cone is touched
				prepareToLookForCone();
				//paramatarize the 95% and greater than 2 sateleites 
				//as consts
			} else if ((_vectorPositionConfidence > 95
				&& _waypointDistanceVector <= dummyDistanceThreshold)
				|| (_gpsPositionConfidence > 16
				&& _waypointDistanceGps <= dummyDistanceThreshold)) {
				//if dummy then must be right on top of dummy 
				//to increment
				_arrivedAtWaypoint = true;
			} else { }
		}

		bool _stopping = false;
		
		//Summary:
		//adjust the setpont of the throttle
		private void adjustThrottleSetpoint() {
			if (!_stopping) {
				//also factor in IR rangefinders here
				platformSpeed = (double)
					(_data.Platform.Data.LeftSonarRaw
					+ _data.Platform.Data.CenterSonarRaw
					+ _data.Platform.Data.RightSonarRaw) / 48.0;
				//expression to make speed inversly porportional to crosstrack error
				navigationSpeed = 100.0 / Math.Pow(_crossTrackErrorVector + 1, 1.0 / 3.0);
				//this is messy but for now maybe it will work
				if (platformSpeed < navigationSpeed)
					speed = platformSpeed;
				else speed = navigationSpeed;
				if (speed > maximumSpeed) speed = maximumSpeed;
				else if (speed < minimumSpeed) speed = minimumSpeed;
				if (Math.Abs(_data.Platform.Data.ThrottleSetpoint
					- _data.Platform.Data.CountsPerSecond)
				< Math.Abs(_data.Platform.Data.ThrottleSetpoint - speed))
					_data.Platform.Data.ThrottleSetpoint = (int)speed;
			} else _data.Platform.Data.ThrottleSetpoint = 0;
		}

		//main "loop" called by navigation controller events
		public void navigateToWaypoint() {
			adjustThrottleSetpoint();
			updateCurrentLocation();
			if (!_stopping) {
				if (!_lookingForCone)
					checkArrivedAtWaypointWaypoint();
				calculateTheta();
				calculateCrossTrackError();
				determineTurnDirection();
				if (!_lookingForCone) {
					if (_turnDirection == "left") {
						steering = _crossTrackErrorVector * steeringKp;
						if (steering > 50) steering = 50;
						_data.Platform.steeringOut((int)Math.Round(steering, 0));
					} else if (_turnDirection == "right") {
						steering = -_crossTrackErrorVector * steeringKp;
						if (steering < -50) steering = -50;
						_data.Platform.steeringOut((int)Math.Round(steering, 0));
					}
				}
			}
			//post update to the UI
			if (updateUiEvent != null)
				updateUiEvent.Invoke();
			if (_pointCount > 0)
				LogData();
		}
		//calculate theta
		//what is the robots desired heading?
		//use vector navigation and gps.

		public void calculateTheta() {
			if (_xDeltaVector == 0 && _yDeltaVector > 0) _thetaVector = 90;
			else if (_xDeltaVector == 0 && _yDeltaVector < 0) _thetaVector = 270;
			else {
				_thetaVector = ConvertDegRad.getDegrees
					(Math.Atan(_yDeltaVector / _xDeltaVector));
				if (_xDeltaVector < 0 && _yDeltaVector > 0)
					_thetaVector = 180.0 + _thetaVector;
				else if (_xDeltaVector < 0 && _yDeltaVector < 0)
					_thetaVector = 180.0 + _thetaVector;
				else if (_xDeltaVector > 0 && _yDeltaVector < 0)
					_thetaVector = 360.0 + _thetaVector; 
			}
			//only look at gps if there is a fix, otherwise bad things happen
			if (_data.Gps.Gga.Data.QualityIndicator > 0) {
				if (_xDeltaGps == 0 && _yDeltaGps > 0) _thetaGps = 90;
				else if (_xDeltaGps == 0 && _yDeltaGps < 0) _thetaGps = 270;
				else {
					_thetaGps = ConvertDegRad.getDegrees
						(Math.Atan(_yDeltaGps / _xDeltaGps));
					if (_xDeltaGps < 0 && _yDeltaGps > 0)
						_thetaGps = 180.0 + _thetaGps;
					else if (_xDeltaGps < 0 && _yDeltaGps < 0)
						_thetaGps = 180.0 + _thetaGps;
					else if (_xDeltaGps > 0 && _yDeltaGps < 0)
						_thetaGps = 360.0 + _thetaGps;
				}
			}
		}

		//the angular difference between the robots heading and its desired heading
		public void calculateCrossTrackError() {
			_crossTrackErrorVector = Math.Abs(_thetaVector - _offsetCompassHeading);
			if (_crossTrackErrorVector > 180)
				_crossTrackErrorVector = 360.0 - _crossTrackErrorVector;
			if (_data.Gps.Gga.Data.QualityIndicator > 0) {
				_crossTrackErrorGps = Math.Abs(_thetaGps - _offsetCompassHeading);
				if (_crossTrackErrorGps > 180)
					_crossTrackErrorGps = 360.0 - _crossTrackErrorGps;
			}
		}

		//based on the robots heading and theta, decide which direction to turn
		public void determineTurnDirection() {
			calculateOffsetCompassHeading();
			if (_thetaVector - _offsetCompassHeading < 0
				&& Math.Abs(_thetaVector - _offsetCompassHeading) >= 180) _turnDirection = "left";
			else if (_thetaVector - _offsetCompassHeading > 0
				&& Math.Abs(_thetaVector - _offsetCompassHeading) >= 180) _turnDirection = "right";
			else if (_thetaVector - _offsetCompassHeading < 0
				&& Math.Abs(_thetaVector - _offsetCompassHeading) < 180) _turnDirection = "right";
			else if (_thetaVector - _offsetCompassHeading > 0
				&& Math.Abs(_thetaVector - _offsetCompassHeading) < 180) _turnDirection = "left";
			else _turnDirection = "null";
		}

		//compensate for the fact that compasses increase in degrees clockwise
		//wheras, trigonometric functions in a coordinate plane go counterclockwise
		//also compensate for thedifference between magnetic north and true north
		public void calculateOffsetCompassHeading() {
			if (_data.Navigation.Data.CompassHeading + 90 <= 180)
				_offsetCompassHeading = 180 
					- (_data.Navigation.Data.CompassHeading + 90.0);
			else _offsetCompassHeading =  360 
				- (_data.Navigation.Data.CompassHeading - 90.0);
		}

		//is the robot near a cone?
		public void prepareToLookForCone() {
			_stopping = true;
			_lookingForCone = true;
			_eventLog.write(DateTime.Now.ToString() + ": prepairing to look for cone, "
				+ "distance to cone is " + _waypointDistanceVector.ToString() + " meters\r\n");
			_coneSearchPatternNumber = 0;
			_spiralTurnOut = -50;
			_data.Vision.initializeCamera();
			_data.Vision.startCamera();
			_data.Vision.updateDataEvent += lookForCone;
		}

		public void lookForCone() {
			++_viewAttempts;
			if (_data.Vision.Data.ConeAquired) {
				++_coneViewCount;
			} else
				_coneViewCount = 0;
			if (_coneViewCount > 8) {
				_data.Vision.updateDataEvent -= lookForCone;
				_viewAttempts = 0;
				_coneViewCount = 0;
				_eventLog.write(DateTime.Now.ToString() + ": can see cone, saving image\r\n");
				_data.Vision.saveRawImage();
				_data.Platform.setManualOverride(true);
				_stopping = false;
				//this triggers the delegate assigned to touchCone
				_data.Navigation.updateDataEvent += touchCone;
			}
			//if dont see cone drive around looking for it
			if (_viewAttempts > 32) {
				//state machine in here to assign 
				//various search patterns to the platform update data event
				_data.Vision.updateDataEvent -= lookForCone;
				elapsedMetersMark = _data.Platform.Data.Meters;
				++_coneSearchPatternNumber;
				_eventLog.write(DateTime.Now.ToString() 
					+ ": can't see cone, executing search pattern " 
					+ _coneSearchPatternNumber.ToString() + "\r\n");
				_data.Vision.saveRawImage();
				_stopping = false;
				switch (_coneSearchPatternNumber) {
					case 1:
						_data.Navigation.updateDataEvent += executeConeSearchPattern1;
						break;
					case 2:
						_data.Navigation.updateDataEvent += executeConeSearchPattern2;
						break;
					case 3:
						_data.Navigation.updateDataEvent += executeConeSearchPattern3;
						break;
					case 4:
						_data.Navigation.updateDataEvent += executeConeSearchPattern4;
						break;
					default:
						_coneSearchPatternNumber = 0;
						break;
				}
			}
		}

		int _coneSearchPatternNumber = 0;
		double elapsedMetersMark = 0;

		private void executeConeSearchPattern1() {
			if (_data.Vision.Data.ConeAquired) {
				_data.Navigation.updateDataEvent -= executeConeSearchPattern1;
				_stopping = true;
				_eventLog.write(DateTime.Now.ToString() + ": spotted cone in search pattern "
				+ "1 resuming looking for cone\r\n");
				_viewAttempts = 0;
				_data.Vision.updateDataEvent += lookForCone;
			} else {
				if (_data.Platform.Data.Meters - elapsedMetersMark > coneDistanceThreshold * 0.5) {
					_data.Navigation.updateDataEvent -= executeConeSearchPattern1;
					_stopping = true;
					_eventLog.write(DateTime.Now.ToString() + ": can't see cone, executing search pattern 2\r\n");
					_data.Vision.saveRawImage();
					elapsedMetersMark = _data.Platform.Data.Meters;
					_stopping = false;
					_data.Navigation.updateDataEvent += executeConeSearchPattern2;
				}
				//drive straight for 1/4 of coneDistanceThreshold
				_data.Platform.steeringOut(0);
			}
		}

		private void executeConeSearchPattern2() {
			if (_data.Vision.Data.ConeAquired) {
				_data.Navigation.updateDataEvent -= executeConeSearchPattern2;
				_stopping = true;
				_eventLog.write(DateTime.Now.ToString() + ": spotted cone in search pattern "
				+ "2 resuming looking for cone\r\n");
				_viewAttempts = 0;
				_data.Vision.updateDataEvent += lookForCone;
			} else {
				if (_data.Platform.Data.Meters - elapsedMetersMark > coneDistanceThreshold * 0.5) {
					_data.Navigation.updateDataEvent -= executeConeSearchPattern2;
					_stopping = true;
					_eventLog.write(DateTime.Now.ToString() + ": can't see cone, executing search pattern 3\r\n");
					_data.Vision.saveRawImage();
					elapsedMetersMark = _data.Platform.Data.Meters;
					_stopping = false;
					_data.Navigation.updateDataEvent += executeConeSearchPattern3;
				}
				//drive hard left for 1/2 of coneDistanceThreshold
				_data.Platform.steeringOut(50);
			}
		}

		private void executeConeSearchPattern3() {
			if (_data.Vision.Data.ConeAquired) {
				_data.Navigation.updateDataEvent -= executeConeSearchPattern3;
				_stopping = true;
				_eventLog.write(DateTime.Now.ToString() + ": spotted cone in search pattern "
				+ "3 resuming looking for cone\r\n");
				_viewAttempts = 0;
				_data.Vision.updateDataEvent += lookForCone;
			} else {
				if (_data.Platform.Data.Meters - elapsedMetersMark > coneDistanceThreshold) {
					_data.Navigation.updateDataEvent -= executeConeSearchPattern3;
					_stopping = true;
					_eventLog.write(DateTime.Now.ToString() + ": can't see cone, executing search pattern 4\r\n");
					_data.Vision.saveRawImage();
					elapsedMetersMark = _data.Platform.Data.Meters;
					_stopping = false;
					_data.Navigation.updateDataEvent += executeConeSearchPattern4;
				}
				//drive hard right for 1/4 of coneDistanceThreshold
				_data.Platform.steeringOut(-50);
			}
		}

		int _spiralTurnOut = -50;
		private void executeConeSearchPattern4() {
			if (_data.Vision.Data.ConeAquired) {
				_data.Navigation.updateDataEvent -= executeConeSearchPattern4;
				_stopping = true;
				_eventLog.write(DateTime.Now.ToString() + ": spotted cone in search pattern "
				+ "4 resuming looking for cone\r\n");
				_viewAttempts = 0;
				_data.Vision.updateDataEvent += lookForCone;
			} else {
				if (_data.Platform.Data.Meters - elapsedMetersMark > coneDistanceThreshold) {
					_data.Navigation.updateDataEvent -= executeConeSearchPattern4;
					_stopping = true;
					_eventLog.write(DateTime.Now.ToString() + ": can't see cone, giving up\r\n");
					_data.Vision.saveRawImage();
					//this is a lie, essentially the robot gives up on the cone here.
					//maybe it should not though, could use GPS to try to find the cone
					//at least use it to plot a new vector because at this point the robot 
					//is lost
					_data.Vision.stopCamera();
					_lookingForCone = false;
					_arrivedAtWaypoint = true;
					_stopping = false;
				}
				//drive in an opening spiral for coneDistanceThreshold
				_data.Platform.steeringOut(++_spiralTurnOut);
			}
		}

		public void touchCone() {
			if (_touchedCone) {
				_data.Navigation.updateDataEvent -= touchCone;
				_arrivedAtWaypoint = true;
				_lookingForCone = false;
				_touchedCone = false;
				_stopping = false;
			} else {
				//assumes the cone has been aquired visually
				if (_data.Platform.Data.LeftBumper
					|| _data.Platform.Data.RightBumper) {
					//the robot touched the cone
					_stopping = true;
					_touchedCone = true;
					_eventLog.write(DateTime.Now.ToString() + ": touched cone\r\n");
					if (_data.Parser.UtmPointSetList[0].Points[_pointCount].Handle != "finish")
						_data.Platform.setManualOverride(false);
					else _eventLog.write(DateTime.Now.ToString() + ": finished race\r\n");
					_eventLog.write(DateTime.Now.ToString() + ": saving image of touched cone\r\n");
					_data.Vision.saveRawImage();
					_data.Vision.stopCamera();
				} else if (_data.Vision.Data.TurnMagnitude != 0) {
					if (!_data.Vision.Data.ConeAquired) {
						++_viewAttempts;
						//reject noise, vision system occasionally looses the cone's signature
						//withought actually loosing sight of it.
						if (_viewAttempts > 4)
							_stopping = true;
						if (_viewAttempts > 32) {
							_data.Navigation.updateDataEvent -= touchCone;
							_eventLog.write(DateTime.Now.ToString() + ": lost sight of cone going to search pattern 2\r\n");
							_data.Platform.setManualOverride(false);
							elapsedMetersMark = _data.Platform.Data.Meters;
							_viewAttempts = 0;
							_coneViewCount = 0;
							_stopping = false;
							_data.Navigation.updateDataEvent += executeConeSearchPattern2;
						}
					} else {
						_stopping = false;
						_viewAttempts = 0;
					}
					//turn in porportion to the error that the cone is from the center of the camera 
					_data.Platform.steeringOut((int)Math.Round((double)_data.Vision.Data.TurnMagnitude * 0.75, 0));
				}
			}
		}

		private void LogData() {
			//write the data out to file
			_dataLog.write(
				DateTime.Now.ToString()
				//race strategy data
				+ "\t" + _pointCount.ToString()
				+ "\t" + _currentLocationVector.Point.X.ToString()
				+ "\t" + _currentLocationVector.Point.Y.ToString()
				+ "\t" + _currentLocationGps.Point.X.ToString()
				+ "\t" + _currentLocationGps.Point.Y.ToString()
				+ "\t" + _waypointDistanceVector.ToString()
				+ "\t" + _waypointDistanceGps.ToString()
				+ "\t" + _crossTrackErrorVector.ToString()
				+ "\t" + _crossTrackErrorGps.ToString()
				+ "\t" + _turnDirection
				+ "\t" + _thetaVector.ToString()
				+ "\t" + _thetaGps.ToString()
				+ "\t" + _lookingForCone.ToString()
				+ "\t" + _coneSearchPatternNumber.ToString()
				+ "\t" + _viewAttempts.ToString()
				+ "\t" + _coneViewCount.ToString()
				+ "\t" + _touchedCone.ToString()
				+ "\t" + _arrivedAtWaypoint.ToString()
				+ "\t" + navigationSpeed.ToString()
				+ "\t" + platformSpeed.ToString()
				+ "\t" + speed.ToString()
				+ "\t" + _stopping.ToString()
				//platform controller data
				+ "\t" + _data.Platform.Data.EncoderCounts.ToString()
				+ "\t" + _data.Platform.Data.Meters.ToString()
				+ "\t" + _data.Platform.Data.CountsPerSecond.ToString()
				+ "\t" + _data.Platform.Data.Speed.ToString()
				+ "\t" + _data.Platform.Data.Throttle.ToString()
				+ "\t" + _data.Platform.Data.Steering.ToString()
				+ "\t" + _data.Platform.Data.Reversing.ToString()
				+ "\t" + _data.Platform.Data.LeftSonar.ToString()
				+ "\t" + _data.Platform.Data.CenterSonar.ToString()
				+ "\t" + _data.Platform.Data.RightSonar.ToString()
				+ "\t" + _data.Platform.Data.LeftBumper.ToString()
				+ "\t" + _data.Platform.Data.RightBumper.ToString()
				+ "\t" + _data.Platform.Data.ManualOverride.ToString()
				+ "\t" + _data.Platform.Data.AutoAvoid.ToString()
				+ "\t" + _data.Platform.Data.IrInstruction.ToString()
				+ "\t" + _data.Platform.Data.Status
				//navigation controller data
				+ "\t" + _data.Navigation.Data.CompassHeading.ToString()
				+ "\t" + _data.Navigation.Data.LeftIrAverage.ToString()
				+ "\t" + _data.Navigation.Data.RightIrAverage.ToString()
				+ "\t" + _data.Navigation.Data.XAccAverage.ToString()
				+ "\t" + _data.Navigation.Data.YAccAverage.ToString()
				+ "\t" + _data.Navigation.Data.ZAccAverage.ToString()
				+ "\t" + _data.Navigation.Data.PlatformBatteryAverage.ToString()
				+ "\t" + _data.Navigation.Data.SystemBatteryAverage.ToString()
				+ "\t" + _data.Navigation.Data.Status.ToString()
				//gps data
				+ "\t" + _data.Gps.Gga.Data.QualityIndicator.ToString()
				+ "\t" + _data.Gps.Gga.Data.SatelitesInView.ToString()
				+ "\t" + _data.Gps.Gga.Data.Latitude.ToString()
				+ "\t" + _data.Gps.Gga.Data.NorthOrSouth.ToString()
				+ "\t" + _data.Gps.Gga.Data.Longitude.ToString()
				+ "\t" + _data.Gps.Gga.Data.EastOrWest.ToString()
				+ "\t" + _data.Gps.Gga.Data.AntennaAltitude.ToString()
				+ "\t" + _data.Gps.Gga.Data.UnitsOfAntennaAltitude.ToString()
				+ "\t" + _data.Gps.Gga.Data.GeoidalSeparation.ToString()
				+ "\t" + _data.Gps.Gga.Data.UnitsOfGeoidalSeperation.ToString()
				+ "\t" + _data.Gps.Gga.Data.HorizontalDilutionOfPrecision.ToString()
				+ "\t" + _data.Gps.Gga.Data.Time.ToString()
				+ "\t" + _data.Gps.Rmc.Data.SpeedOverGround.ToString()
				+ "\t" + _data.Gps.Rmc.Data.TrackMadeGood.ToString()
				+ "\t" + _data.Gps.Rmc.Data.MagneticVariation.ToString()
				+ "\t" + _data.Gps.Rmc.Data.EastOrWestM.ToString() + "\r\n");
		}
    }
}
